@@ -1,3 +1,4 @@
+{-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE NamedFieldPuns #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards #-}
@@ -15,6 +16,7 @@ import Foreign.Ptr (minusPtr, plusPtr)
 import Network.ByteOrder
 import Network.HTTP.Semantics.Client
 import Network.HTTP.Semantics.IO
+import System.ThreadManager
 
 import Imports
 import Network.HPACK (setLimitForEncoding, toTokenHeaderTable)
@@ -36,14 +38,6 @@ data Switch
     | O Output
     | Flush
 
-wrapException :: E.SomeException -> IO ()
-wrapException se
-    | isAsyncException se = E.throwIO se
-    | Just GoAwayIsSent <- E.fromException se = return ()
-    | Just ConnectionIsClosed <- E.fromException se = return ()
-    | Just (e :: HTTP2Error) <- E.fromException se = E.throwIO e
-    | otherwise = E.throwIO $ BadThingHappen se
-
 -- Peer SETTINGS_INITIAL_WINDOW_SIZE
 -- Adjusting initial window size for streams
 updatePeerSettings :: Context -> SettingsList -> IO ()
@@ -65,21 +59,38 @@ updatePeerSettings Context{peerSettings, oddStreamTable, evenStreamTable} peerAl
     updateAllStreamTxFlow siz strms =
         forM_ strms $ \strm -> increaseStreamWindowSize strm siz
 
+checkDone :: Context -> Int -> IO Bool
+checkDone Context{..} 0 = atomically $ do
+    isEmptyC <- isEmptyTQueue controlQ
+    isEmptyO <- isEmptyTQueue outputQ
+    if not isEmptyC || not isEmptyO
+        then
+            return False
+        else do
+            gone <- isAllGone threadManager
+            unless gone retry
+            done <- readTVar receiverDone
+            unless done retry
+            return True
+checkDone _ _ = return False
+
 frameSender :: Context -> Config -> IO ()
 frameSender
-    ctx@Context{outputQ, controlQ, encodeDynamicTable, outputBufferLimit, senderDone}
+    ctx@Context{outputQ, controlQ, encodeDynamicTable, outputBufferLimit}
     Config{..} = do
         labelMe "H2 sender"
-        (loop 0 `E.finally` setSenderDone) `E.catch` wrapException
+        loop 0
       where
         ----------------------------------------------------------------
         loop :: Offset -> IO ()
         loop off = do
-            x <- atomically $ dequeue off
-            case x of
-                C ctl -> flushN off >> control ctl >> loop 0
-                O out -> outputAndSync out off >>= flushIfNecessary >>= loop
-                Flush -> flushN off >> loop 0
+            done <- checkDone ctx off
+            unless done $ do
+                x <- atomically $ dequeue off
+                case x of
+                    C ctl -> flushN off >> control ctl >> loop 0
+                    O out -> outputAndSync out off >>= flushIfNecessary >>= loop
+                    Flush -> flushN off >> loop 0
 
         -- Flush the connection buffer to the socket, where the first 'n' bytes of
         -- the buffer are filled.
@@ -115,7 +126,6 @@ frameSender
 
         -- called with off == 0
         control :: Control -> IO ()
-        control (CFinish e) = E.throwIO e
         control (CFrames ms xs) = do
             buf <- copyAll xs confWriteBuffer
             let off = buf `minusPtr` confWriteBuffer
@@ -163,10 +173,12 @@ frameSender
                         return off'
 
         resetStream :: Stream -> ErrorCode -> E.SomeException -> IO ()
-        resetStream strm err e = do
-            closed ctx strm (ResetByMe e)
-            let rst = resetFrame err $ streamNumber strm
-            enqueueControl controlQ $ CFrames Nothing [rst]
+        resetStream strm err e
+            | isAsyncException e = E.throwIO e
+            | otherwise = do
+                closed ctx strm (ResetByMe e)
+                let rst = resetFrame err $ streamNumber strm
+                enqueueControl controlQ $ CFrames Nothing [rst]
 
         ----------------------------------------------------------------
         outputHeader
@@ -203,35 +215,34 @@ frameSender
             let payloadOff = off0 + frameHeaderLength
                 datBuf = confWriteBuffer `plusPtr` payloadOff
                 datBufSiz = buflim - payloadOff
-            curr datBuf (min datBufSiz lim) >>= \next ->
-                case next of
-                    Next datPayloadLen reqflush mnext -> do
-                        NextTrailersMaker tlrmkr' <- runTrailersMaker tlrmkr datBuf datPayloadLen
-                        fillDataHeader
-                            strm
-                            off0
-                            datPayloadLen
-                            mnext
-                            tlrmkr'
-                            out
-                            reqflush
-                    CancelNext mErr -> do
-                        -- Stream cancelled
-                        --
-                        -- At this point, the headers have already been sent.
-                        -- Therefore, the stream cannot be in the 'Idle' state, so we
-                        -- are justified in sending @RST_STREAM@.
-                        --
-                        -- By the invariant on the 'outputQ', there are no other
-                        -- outputs for this stream already enqueued. Therefore, we can
-                        -- safely cancel it knowing that we won't try and send any
-                        -- more data frames on this stream.
-                        case mErr of
-                            Just err ->
-                                resetStream strm InternalError err
-                            Nothing ->
-                                resetStream strm Cancel (E.toException CancelledStream)
-                        return (off0, Nothing)
+            curr datBuf (min datBufSiz lim) >>= \case
+                Next datPayloadLen reqflush mnext -> do
+                    NextTrailersMaker tlrmkr' <- runTrailersMaker tlrmkr datBuf datPayloadLen
+                    fillDataHeader
+                        strm
+                        off0
+                        datPayloadLen
+                        mnext
+                        tlrmkr'
+                        out
+                        reqflush
+                CancelNext mErr -> do
+                    -- Stream cancelled
+                    --
+                    -- At this point, the headers have already been sent.
+                    -- Therefore, the stream cannot be in the 'Idle' state, so we
+                    -- are justified in sending @RST_STREAM@.
+                    --
+                    -- By the invariant on the 'outputQ', there are no other
+                    -- outputs for this stream already enqueued. Therefore, we can
+                    -- safely cancel it knowing that we won't try and send any
+                    -- more data frames on this stream.
+                    case mErr of
+                        Just err ->
+                            resetStream strm InternalError err
+                        Nothing ->
+                            resetStream strm Cancel (E.toException CancelledStream)
+                    return (off0, Nothing)
         output (Output strm (OPush ths pid) _) off0 _lim = do
             -- Creating a push promise header
             -- Frame id should be associated stream id from the client.
@@ -260,7 +271,7 @@ frameSender
           where
             eos = if endOfStream then setEndStream else id
             getFlag [] = eos $ setEndHeader defaultFlags
-            getFlag _ = eos $ defaultFlags
+            getFlag _ = eos defaultFlags
 
             continue :: Offset -> TokenHeaderList -> FrameType -> IO Offset
             continue off [] _ = return off
@@ -388,5 +399,3 @@ frameSender
                     , flags = flag
                     , streamId = sid
                     }
-
-        setSenderDone = atomically $ writeTVar senderDone True
