@@ -2,7 +2,6 @@
 {-# LANGUAGE DerivingStrategies #-}
 {-# LANGUAGE NamedFieldPuns #-}
 {-# LANGUAGE RankNTypes #-}
-{-# LANGUAGE RecordWildCards #-}
 
 module Network.HTTP2.H2.Stream where
 
@@ -102,10 +101,7 @@ closeAllStreams ovar evar mErr' = do
             mErr'
 
     err :: Either SomeException a
-    err =
-        Left $
-            fromMaybe (toException ConnectionIsClosed) $
-                mErr
+    err = Left $ fromMaybe (toException ConnectionIsClosed) mErr
 
 ----------------------------------------------------------------
 
@@ -125,11 +121,7 @@ withOutBodyIface tbq unmask k = do
     terminated <- newTVarIO Nothing
     let whenNotTerminated act = do
             mTerminated <- readTVar terminated
-            case mTerminated of
-                Just reason ->
-                    throwSTM reason
-                Nothing ->
-                    act
+            maybe act throwSTM mTerminated
 
         terminateWith reason act = do
             mTerminated <- readTVar terminated
@@ -158,10 +150,11 @@ withOutBodyIface tbq unmask k = do
                     atomically $
                         whenNotTerminated $
                             writeTBQueue tbq StreamingFlush
-                , outBodyCancel = \mErr ->
-                    atomically $
-                        terminateWith StreamCancelled $
-                            writeTBQueue tbq (StreamingCancelled mErr)
+                , outBodyCancel =
+                    atomically
+                        . terminateWith StreamCancelled
+                        . writeTBQueue tbq
+                        . StreamingCancelled
                 }
         finished = atomically $ do
             terminateWith StreamOutOfScope $
