@@ -1,5 +1,5 @@
 {-# LANGUAGE OverloadedStrings #-}
-{-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE RankNTypes #-}
 
 module Main where
 
@@ -9,6 +9,7 @@ import Control.Monad
 import qualified Data.ByteString.Char8 as C8
 import Network.HTTP2.Client
 import Network.Run.TCP (runTCPClient)
+import Network.Socket
 import System.Console.GetOpt
 import System.Environment
 import System.Exit
@@ -22,6 +23,7 @@ defaultOptions =
         { optPerformance = 0
         , optNumOfReqs = 1
         , optMonitor = False
+        , optInteractive = False
         }
 
 usage :: String
@@ -44,6 +46,11 @@ options =
         ["monitor"]
         (NoArg (\opts -> opts{optMonitor = True}))
         "run thread monitor"
+    , Option
+        ['i']
+        ["interactive"]
+        (NoArg (\o -> o{optInteractive = True}))
+        "enter interactive mode"
     ]
 
 showUsageAndExit :: String -> IO a
@@ -70,8 +77,19 @@ main = do
         h : p : ps -> return (h, p, C8.pack <$> ps)
     when (optMonitor opts) $ void $ forkIO $ monitor $ threadDelay 1000000
     let cliconf = defaultClientConfig{authority = host}
-    runTCPClient host port $ \s ->
-        E.bracket
-            (allocSimpleConfig' s 4096 5000000)
-            freeSimpleConfig
-            (\conf -> run cliconf conf $ client opts paths)
+    run' cliconf host port $ client' opts paths
+
+run' :: ClientConfig -> HostName -> ServiceName -> Client a -> IO a
+run' cliconf host port f = runTCPClient host port $ \s ->
+    E.bracket
+        (allocSimpleConfig' s 4096 10000000)
+        freeSimpleConfig
+        (\conf -> run cliconf conf f)
+
+client' :: Options -> [Path] -> Client ()
+client' opts paths sendRequest _aux
+    | optInteractive opts = do
+        let action = client opts paths sendRequest _aux
+        console opts paths action _aux
+        return ()
+    | otherwise = client opts paths sendRequest _aux
